abs_file_name() believed that a file name that begins with / or contains :
is absolute and that any other file name is relative.  On Windows, this is
wrong in at least the following ways:

   * / and \ are interchangeable on Windows.

   * A name that begins with \\ or // is also absolute.

   * A name that begins with X: but not X:\ is not absolute.

   * A name with : in some position other than the second position is
     not absolute (although it might not be valid either?).

Furthermore, Windows has more than one current working directory (one per
volume letter), so trying to make a file name absolute by just prefixing
the current working directory for the current volume results in silliness.

This patch attempts to fix the problem.

Found by inspection.

CC: Alin Gabriel Serdean <aserd...@ovn.org>
Signed-off-by: Ben Pfaff <b...@ovn.org>
---
 lib/util.c | 58 +++++++++++++++++++++++++++++++++++++++-------------------
 lib/util.h |  1 +
 2 files changed, 40 insertions(+), 19 deletions(-)

diff --git a/lib/util.h b/lib/util.h
index 0061bb986520..fd414489c7fe 100644
--- a/lib/util.h
+++ b/lib/util.h
@@ -231,6 +231,7 @@ char *dir_name(const char *file_name);
 char *base_name(const char *file_name);
 #endif
 char *abs_file_name(const char *dir, const char *file_name);
+bool is_file_name_absolute(const char *);
 
 char *follow_symlinks(const char *filename);
 
diff --git a/lib/util.c b/lib/util.c
index 7152b55392be..d3c62988bf83 100644
--- a/lib/util.c
+++ b/lib/util.c
@@ -1049,37 +1049,57 @@ base_name(const char *file_name)
 }
 #endif /* _WIN32 */
 
-/* If 'file_name' starts with '/', returns a copy of 'file_name'.  Otherwise,
+bool
+is_file_name_absolute(const char *fn)
+{
+#ifdef _WIN32
+    /* An absolute path begins with X:\ or \\. */
+    return ((fn[0] && fn[1] == ':' && strchr("/\\", fn[2]))
+            || (strchr("/\\", fn[0]) && strchr("/\\", fn[1])));
+#else
+    /* An absolute path begins with /. */
+    return fn[0] == '/';
+#endif
+}
+
+/* If 'file_name' is absolute, returns a copy of 'file_name'.  Otherwise,
  * returns an absolute path to 'file_name' considering it relative to 'dir',
  * which itself must be absolute.  'dir' may be null or the empty string, in
  * which case the current working directory is used.
  *
- * Additionally on Windows, if 'file_name' has a ':', returns a copy of
- * 'file_name'
- *
  * Returns a null pointer if 'dir' is null and getcwd() fails. */
 char *
 abs_file_name(const char *dir, const char *file_name)
 {
-    if (file_name[0] == '/') {
+    /* If it's already absolute, return a copy. */
+    if (is_file_name_absolute(file_name)) {
         return xstrdup(file_name);
-#ifdef _WIN32
-    } else if (strchr(file_name, ':')) {
-        return xstrdup(file_name);
-#endif
-    } else if (dir && dir[0]) {
+    }
+
+    /* If a base dir was supplied, use it.  We assume, without checking, that
+     * the base dir is absolute.*/
+    if (dir && dir[0]) {
         char *separator = dir[strlen(dir) - 1] == '/' ? "" : "/";
         return xasprintf("%s%s%s", dir, separator, file_name);
-    } else {
-        char *cwd = get_cwd();
-        if (cwd) {
-            char *abs_name = xasprintf("%s/%s", cwd, file_name);
-            free(cwd);
-            return abs_name;
-        } else {
-            return NULL;
-        }
     }
+
+#if _WIN32
+    /* It's a little complicated to make an absolute path on Windows because a
+     * relative path might still specify a drive letter.  The OS has a function
+     * to do the job for us, so use it. */
+    char abs_path[MAX_PATH];
+    DWORD n = GetFullPathName(file_name, sizeof abs_path, abs_path, NULL);
+    return n > 0 && n <= sizeof abs_path ? xmemdup0(abs_path, n) : NULL;
+#else
+    /* Outside Windows, do the job ourselves. */
+    char *cwd = get_cwd();
+    if (!cwd) {
+        return NULL;
+    }
+    char *abs_name = xasprintf("%s/%s", cwd, file_name);
+    free(cwd);
+    return abs_name;
+#endif
 }
 
 /* Like readlink(), but returns the link name as a null-terminated string in
-- 
2.16.1

_______________________________________________
dev mailing list
d...@openvswitch.org
https://mail.openvswitch.org/mailman/listinfo/ovs-dev

Reply via email to